@@ -6,8 +6,9 @@
     :license: BSD, see LICENSE for more details.
 """
 
+from __future__ import annotations
+
 import bisect
-import contextlib
 import copy
 import datetime
 import functools
@@ -17,13 +18,27 @@ import numbers
 import operator
 import re
 import warnings
-from typing import List
-
-from packaging import version
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Generic,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    overload,
+)
 
+from ._typing import S, Shape, UnitLike, _MagnitudeType
 from .compat import (
-    HAS_NUMPY_ARRAY_FUNCTION,
-    NUMPY_VER,
+    HAS_NUMPY,
     _to_magnitude,
     babel_parse,
     compute,
@@ -46,6 +61,7 @@ from .errors import (
 )
 from .formatting import (
     _pretty_fmt_exponent,
+    extract_custom_flags,
     ndarray_to_latex,
     remove_custom_flags,
     siunitx_format_unit,
@@ -70,6 +86,14 @@ from .util import (
     to_units_container,
 )
 
+if TYPE_CHECKING:
+    from . import Context, Unit
+    from .registry import BaseRegistry
+    from .unit import UnitsContainer as UnitsContainerT
+
+    if HAS_NUMPY:
+        import numpy as np  # noqa
+
 
 class _Exception(Exception):  # pragma: no cover
     def __init__(self, internal):
@@ -143,21 +167,11 @@ def check_dask_array(f):
     return wrapper
 
 
-@contextlib.contextmanager
-def printoptions(*args, **kwargs):
-    """Numpy printoptions context manager released with version 1.15.0
-    https://docs.scipy.org/doc/numpy/reference/generated/numpy.printoptions.html
-    """
-
-    opts = np.get_printoptions()
-    try:
-        np.set_printoptions(*args, **kwargs)
-        yield np.get_printoptions()
-    finally:
-        np.set_printoptions(**opts)
+# Workaround to bypass dynamically generated Quantity with overload method
+Magnitude = TypeVar("Magnitude")
 
 
-class Quantity(PrettyIPython, SharedRegistryObject):
+class Quantity(PrettyIPython, SharedRegistryObject, Generic[_MagnitudeType]):
     """Implements a class to describe a physical quantity:
     the product of a numerical value and a unit of measurement.
 
@@ -174,21 +188,22 @@ class Quantity(PrettyIPython, SharedRegistryObject):
     """
 
     #: Default formatting string.
-    default_format = ""
+    default_format: str = ""
+    _magnitude: _MagnitudeType
 
     @property
-    def force_ndarray(self):
+    def force_ndarray(self) -> bool:
         return self._REGISTRY.force_ndarray
 
     @property
-    def force_ndarray_like(self):
+    def force_ndarray_like(self) -> bool:
         return self._REGISTRY.force_ndarray_like
 
     @property
-    def UnitsContainer(self):
+    def UnitsContainer(self) -> Callable[..., UnitsContainerT]:
         return self._REGISTRY.UnitsContainer
 
-    def __reduce__(self):
+    def __reduce__(self) -> tuple:
         """Allow pickling quantities. Since UnitRegistries are not pickled, upon
         unpickling the new object is always attached to the application registry.
         """
@@ -198,55 +213,76 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         # build_quantity_class can't be pickled
         return _unpickle_quantity, (Quantity, self.magnitude, self._units)
 
+    @overload
+    def __new__(
+        cls, value: str, units: Optional[UnitLike] = None
+    ) -> Quantity[Magnitude]:
+        ...
+
+    @overload
+    def __new__(  # type: ignore[misc]
+        cls, value: Sequence, units: Optional[UnitLike] = None
+    ) -> Quantity[np.ndarray]:
+        ...
+
+    @overload
+    def __new__(
+        cls, value: Quantity[Magnitude], units: Optional[UnitLike] = None
+    ) -> Quantity[Magnitude]:
+        ...
+
+    @overload
+    def __new__(
+        cls, value: Magnitude, units: Optional[UnitLike] = None
+    ) -> Quantity[Magnitude]:
+        ...
+
     def __new__(cls, value, units=None):
         if is_upcast_type(type(value)):
             raise TypeError(f"Quantity cannot wrap upcast type {type(value)}")
-        elif units is None:
-            if isinstance(value, str):
-                if value == "":
-                    raise ValueError(
-                        "Expression to parse as Quantity cannot " "be an empty string."
-                    )
-                ureg = SharedRegistryObject.__new__(cls)._REGISTRY
-                inst = ureg.parse_expression(value)
-                return cls.__new__(cls, inst)
-            elif isinstance(value, cls):
-                inst = copy.copy(value)
-            else:
-                inst = SharedRegistryObject.__new__(cls)
-                inst._magnitude = _to_magnitude(
-                    value, inst.force_ndarray, inst.force_ndarray_like
-                )
-                inst._units = inst.UnitsContainer()
-        elif isinstance(units, (UnitsContainer, UnitDefinition)):
-            inst = SharedRegistryObject.__new__(cls)
-            inst._magnitude = _to_magnitude(
-                value, inst.force_ndarray, inst.force_ndarray_like
-            )
-            inst._units = units
-        elif isinstance(units, str):
-            inst = SharedRegistryObject.__new__(cls)
-            inst._magnitude = _to_magnitude(
-                value, inst.force_ndarray, inst.force_ndarray_like
-            )
-            inst._units = inst._REGISTRY.parse_units(units)._units
-        elif isinstance(units, SharedRegistryObject):
-            if isinstance(units, Quantity) and units.magnitude != 1:
-                inst = copy.copy(units)
-                logger.warning(
-                    "Creating new Quantity using a non unity " "Quantity as units."
-                )
-            else:
-                inst = SharedRegistryObject.__new__(cls)
-                inst._units = units._units
-            inst._magnitude = _to_magnitude(
-                value, inst.force_ndarray, inst.force_ndarray_like
+
+        if units is None and isinstance(value, str) and value == "":
+            raise ValueError(
+                "Expression to parse as Quantity cannot be an empty string."
             )
+
+        if units is None and isinstance(value, str):
+            ureg = SharedRegistryObject.__new__(cls)._REGISTRY
+            inst = ureg.parse_expression(value)
+            return cls.__new__(cls, inst)
+
+        if units is None and isinstance(value, cls):
+            return copy.copy(value)
+
+        inst = SharedRegistryObject().__new__(cls)
+        if units is None:
+            units = inst.UnitsContainer()
         else:
-            raise TypeError(
-                "units must be of type str, Quantity or "
-                "UnitsContainer; not {}.".format(type(units))
+            if isinstance(units, (UnitsContainer, UnitDefinition)):
+                units = units
+            elif isinstance(units, str):
+                units = inst._REGISTRY.parse_units(units)._units
+            elif isinstance(units, SharedRegistryObject):
+                if isinstance(units, Quantity) and units.magnitude != 1:
+                    units = copy.copy(units)._units
+                    logger.warning(
+                        "Creating new Quantity using a non unity Quantity as units."
+                    )
+                else:
+                    units = units._units
+            else:
+                raise TypeError(
+                    "units must be of type str, Quantity or "
+                    "UnitsContainer; not {}.".format(type(units))
+                )
+        if isinstance(value, cls):
+            magnitude = value.to(units)._magnitude
+        else:
+            magnitude = _to_magnitude(
+                value, inst.force_ndarray, inst.force_ndarray_like
             )
+        inst._magnitude = magnitude
+        inst._units = units
 
         inst.__used = False
         inst.__handling = None
@@ -257,7 +293,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
     def debug_used(self):
         return self.__used
 
-    def __iter__(self):
+    def __iter__(self: Quantity[Iterable[S]]) -> Iterator[S]:
         # Make sure that, if self.magnitude is not iterable, we raise TypeError as soon
         # as one calls iter(self) without waiting for the first element to be drawn from
         # the iterator
@@ -269,34 +305,34 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return it_outer()
 
-    def __copy__(self):
+    def __copy__(self) -> Quantity[_MagnitudeType]:
         ret = self.__class__(copy.copy(self._magnitude), self._units)
         ret.__used = self.__used
         return ret
 
-    def __deepcopy__(self, memo):
+    def __deepcopy__(self, memo) -> Quantity[_MagnitudeType]:
         ret = self.__class__(
             copy.deepcopy(self._magnitude, memo), copy.deepcopy(self._units, memo)
         )
         ret.__used = self.__used
         return ret
 
-    def __str__(self):
+    def __str__(self) -> str:
         if self._REGISTRY.fmt_locale is not None:
             return self.format_babel()
 
         return format(self)
 
-    def __bytes__(self):
+    def __bytes__(self) -> bytes:
         return str(self).encode(locale.getpreferredencoding())
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         if isinstance(self._magnitude, float):
             return f"<Quantity({self._magnitude:.9}, '{self._units}')>"
         else:
             return f"<Quantity({self._magnitude}, '{self._units}')>"
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         self_base = self.to_base_units()
         if self_base.dimensionless:
             return hash(self_base.magnitude)
@@ -305,22 +341,51 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
     _exp_pattern = re.compile(r"([0-9]\.?[0-9]*)e(-?)\+?0*([0-9]+)")
 
-    def __format__(self, spec):
+    def __format__(self, spec: str) -> str:
         if self._REGISTRY.fmt_locale is not None:
             return self.format_babel(spec)
 
-        spec = spec or self.default_format
+        mspec = remove_custom_flags(spec)
+        uspec = extract_custom_flags(spec)
+
+        default_mspec = remove_custom_flags(self.default_format)
+        default_uspec = extract_custom_flags(self.default_format)
+        if spec:
+            if not uspec and default_uspec:
+                warnings.warn(
+                    (
+                        "The given format spec does not contain a unit formatter."
+                        " Falling back to the builtin defaults, but in the future"
+                        " the unit formatter specified in the `default_format`"
+                        " attribute will be used instead."
+                    ),
+                    DeprecationWarning,
+                )
+            if not mspec and default_mspec:
+                warnings.warn(
+                    (
+                        "The given format spec does not contain a magnitude formatter."
+                        " Falling back to the builtin defaults, but in the future"
+                        " the magnitude formatter specified in the `default_format`"
+                        " attribute will be used instead."
+                    ),
+                    DeprecationWarning,
+                )
+        else:
+            mspec, uspec = default_mspec, default_uspec
 
         # If Compact is selected, do it at the beginning
         if "#" in spec:
-            spec = spec.replace("#", "")
+            # TODO: don't replace '#'
+            mspec = mspec.replace("#", "")
+            uspec = uspec.replace("#", "")
             obj = self.to_compact()
         else:
             obj = self
 
-        if "L" in spec:
+        if "L" in uspec:
             allf = plain_allf = r"{}\ {}"
-        elif "H" in spec:
+        elif "H" in uspec:
             allf = plain_allf = "{} {}"
             if iterable(obj.magnitude):
                 # Use HTML table instead of plain text template for array-likes
@@ -334,19 +399,19 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         else:
             allf = plain_allf = "{} {}"
 
-        if "Lx" in spec:
+        if "Lx" in uspec:
             # the LaTeX siunitx code
-            spec = spec.replace("Lx", "")
             # TODO: add support for extracting options
             opts = ""
-            ustr = siunitx_format_unit(obj.units)
+            ustr = siunitx_format_unit(obj.units._units, obj._REGISTRY)
             allf = r"\SI[%s]{{{}}}{{{}}}" % opts
         else:
             # Hand off to unit formatting
-            ustr = format(obj.units, spec)
+            # TODO: only use `uspec` after completing the deprecation cycle
+            ustr = format(obj.units, mspec + uspec)
 
-        mspec = remove_custom_flags(spec)
-        if "H" in spec:
+        # mspec = remove_custom_flags(spec)
+        if "H" in uspec:
             # HTML formatting
             if hasattr(obj.magnitude, "_repr_html_"):
                 # If magnitude has an HTML repr, nest it within Pint's
@@ -355,12 +420,20 @@ class Quantity(PrettyIPython, SharedRegistryObject):
                 if isinstance(self.magnitude, ndarray):
                     # Use custom ndarray text formatting with monospace font
                     formatter = "{{:{}}}".format(mspec)
-                    with printoptions(formatter={"float_kind": formatter.format}):
-                        mstr = (
-                            "<pre>"
-                            + format(obj.magnitude).replace("\n", "<br>")
-                            + "</pre>"
-                        )
+                    # Need to override for scalars, which are detected as iterable,
+                    # and don't respond to printoptions.
+                    if self.magnitude.ndim == 0:
+                        allf = plain_allf = "{} {}"
+                        mstr = formatter.format(obj.magnitude)
+                    else:
+                        with np.printoptions(
+                            formatter={"float_kind": formatter.format}
+                        ):
+                            mstr = (
+                                "<pre>"
+                                + format(obj.magnitude).replace("\n", "<br>")
+                                + "</pre>"
+                            )
                 elif not iterable(obj.magnitude):
                     # Use plain text for scalars
                     mstr = format(obj.magnitude, mspec)
@@ -372,23 +445,27 @@ class Quantity(PrettyIPython, SharedRegistryObject):
                         + "</pre>"
                     )
         elif isinstance(self.magnitude, ndarray):
-            if "L" in spec:
+            if "L" in uspec:
                 # Use ndarray LaTeX special formatting
                 mstr = ndarray_to_latex(obj.magnitude, mspec)
             else:
-                # Use custom ndarray text formatting
+                # Use custom ndarray text formatting--need to handle scalars differently
+                # since they don't respond to printoptions
                 formatter = "{{:{}}}".format(mspec)
-                with printoptions(formatter={"float_kind": formatter.format}):
-                    mstr = format(obj.magnitude).replace("\n", "")
+                if obj.magnitude.ndim == 0:
+                    mstr = formatter.format(obj.magnitude)
+                else:
+                    with np.printoptions(formatter={"float_kind": formatter.format}):
+                        mstr = format(obj.magnitude).replace("\n", "")
         else:
             mstr = format(obj.magnitude, mspec).replace("\n", "")
 
-        if "L" in spec:
+        if "L" in uspec:
             mstr = self._exp_pattern.sub(r"\1\\times 10^{\2\3}", mstr)
-        elif "H" in spec or "P" in spec:
+        elif "H" in uspec or "P" in uspec:
             m = self._exp_pattern.match(mstr)
             _exp_formatter = (
-                _pretty_fmt_exponent if "P" in spec else lambda s: f"<sup>{s}</sup>"
+                _pretty_fmt_exponent if "P" in uspec else lambda s: f"<sup>{s}</sup>"
             )
             if m:
                 exp = int(m.group(2) + m.group(3))
@@ -407,7 +484,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             p.text(" ")
             p.pretty(self.units)
 
-    def format_babel(self, spec="", **kwspec):
+    def format_babel(self, spec: str = "", **kwspec: Any) -> str:
         spec = spec or self.default_format
 
         # standard cases
@@ -432,16 +509,16 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         ).replace("\n", "")
 
     @property
-    def magnitude(self):
+    def magnitude(self) -> _MagnitudeType:
         """Quantity's magnitude. Long form for `m`"""
         return self._magnitude
 
     @property
-    def m(self):
+    def m(self) -> _MagnitudeType:
         """Quantity's magnitude. Short form for `magnitude`"""
         return self._magnitude
 
-    def m_as(self, units):
+    def m_as(self, units) -> _MagnitudeType:
         """Quantity's magnitude expressed in particular units.
 
         Parameters
@@ -456,31 +533,31 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         return self.to(units).magnitude
 
     @property
-    def units(self):
+    def units(self) -> "Unit":
         """Quantity's units. Long form for `u`"""
         return self._REGISTRY.Unit(self._units)
 
     @property
-    def u(self):
+    def u(self) -> "Unit":
         """Quantity's units. Short form for `units`"""
         return self._REGISTRY.Unit(self._units)
 
     @property
-    def unitless(self):
+    def unitless(self) -> bool:
         """ """
         return not bool(self.to_root_units()._units)
 
     @property
-    def dimensionless(self):
+    def dimensionless(self) -> bool:
         """ """
         tmp = self.to_root_units()
 
         return not bool(tmp.dimensionality)
 
-    _dimensionality = None
+    _dimensionality: Optional[UnitsContainerT] = None
 
     @property
-    def dimensionality(self):
+    def dimensionality(self) -> UnitsContainerT:
         """
         Returns
         -------
@@ -492,12 +569,12 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return self._dimensionality
 
-    def check(self, dimension):
+    def check(self, dimension: UnitLike) -> bool:
         """Return true if the quantity's dimension matches passed dimension."""
         return self.dimensionality == self._REGISTRY.get_dimensionality(dimension)
 
     @classmethod
-    def from_list(cls, quant_list, units=None):
+    def from_list(cls, quant_list: List[Quantity], units=None) -> Quantity[np.ndarray]:
         """Transforms a list of Quantities into an numpy.array quantity.
         If no units are specified, the unit of the first element will be used.
         Same as from_sequence.
@@ -519,7 +596,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         return cls.from_sequence(quant_list, units=units)
 
     @classmethod
-    def from_sequence(cls, seq, units=None):
+    def from_sequence(cls, seq: Sequence[Quantity], units=None) -> Quantity[np.ndarray]:
         """Transforms a sequence of Quantities into an numpy.array quantity.
         If no units are specified, the unit of the first element will be used.
 
@@ -557,7 +634,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
     def from_tuple(cls, tup):
         return cls(tup[0], cls._REGISTRY.UnitsContainer(tup[1]))
 
-    def to_tuple(self):
+    def to_tuple(self) -> Tuple[_MagnitudeType, Tuple[Tuple[str]]]:
         return self.m, tuple(self._units.items())
 
     def compatible_units(self, *contexts):
@@ -567,7 +644,9 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return self._REGISTRY.get_compatible_units(self._units)
 
-    def is_compatible_with(self, other, *contexts, **ctx_kwargs):
+    def is_compatible_with(
+        self, other: Any, *contexts: Union[str, Context], **ctx_kwargs: Any
+    ) -> bool:
         """check if the other object is compatible
 
         Parameters
@@ -584,14 +663,16 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         -------
         bool
         """
-        if contexts:
+        from .unit import Unit
+
+        if contexts or self._REGISTRY._active_ctx:
             try:
                 self.to(other, *contexts, **ctx_kwargs)
                 return True
             except DimensionalityError:
                 return False
 
-        if isinstance(other, (self._REGISTRY.Quantity, self._REGISTRY.Unit)):
+        if isinstance(other, (Quantity, Unit)):
             return self.dimensionality == other.dimensionality
 
         if isinstance(other, str):
@@ -620,7 +701,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             inplace=is_duck_array_type(type(self._magnitude)),
         )
 
-    def ito(self, other=None, *contexts, **ctx_kwargs):
+    def ito(self, other=None, *contexts, **ctx_kwargs) -> None:
         """Inplace rescale to different units.
 
         Parameters
@@ -639,7 +720,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return None
 
-    def to(self, other=None, *contexts, **ctx_kwargs):
+    def to(self, other=None, *contexts, **ctx_kwargs) -> Quantity[_MagnitudeType]:
         """Return Quantity rescaled to different units.
 
         Parameters
@@ -661,7 +742,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return self.__class__(magnitude, other)
 
-    def ito_root_units(self):
+    def ito_root_units(self) -> None:
         """Return Quantity rescaled to root units."""
 
         _, other = self._REGISTRY._get_root_units(self._units)
@@ -671,7 +752,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return None
 
-    def to_root_units(self):
+    def to_root_units(self) -> Quantity[_MagnitudeType]:
         """Return Quantity rescaled to root units."""
 
         _, other = self._REGISTRY._get_root_units(self._units)
@@ -680,7 +761,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return self.__class__(magnitude, other)
 
-    def ito_base_units(self):
+    def ito_base_units(self) -> None:
         """Return Quantity rescaled to base units."""
 
         _, other = self._REGISTRY._get_base_units(self._units)
@@ -690,7 +771,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return None
 
-    def to_base_units(self):
+    def to_base_units(self) -> Quantity[_MagnitudeType]:
         """Return Quantity rescaled to base units."""
 
         _, other = self._REGISTRY._get_base_units(self._units)
@@ -699,7 +780,24 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return self.__class__(magnitude, other)
 
-    def ito_reduced_units(self):
+    def _get_reduced_units(self, units):
+        # loop through individual units and compare to each other unit
+        # can we do better than a nested loop here?
+        for unit1, exp in units.items():
+            # make sure it wasn't already reduced to zero exponent on prior pass
+            if unit1 not in units:
+                continue
+            for unit2 in units:
+                # get exponent after reduction
+                exp = units[unit1]
+                if unit1 != unit2:
+                    power = self._REGISTRY._get_dimensionality_ratio(unit1, unit2)
+                    if power:
+                        units = units.add(unit2, exp / power).remove([unit1])
+                        break
+        return units
+
+    def ito_reduced_units(self) -> None:
         """Return Quantity scaled in place to reduced units, i.e. one unit per
         dimension. This will not reduce compound units (e.g., 'J/kg' will not
         be reduced to m**2/s**2), nor can it make use of contexts at this time.
@@ -711,41 +809,36 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         if len(self._units) == 1:
             return None
 
-        newunits = self._units.copy()
-        # loop through individual units and compare to each other unit
-        # can we do better than a nested loop here?
-        for unit1, exp in self._units.items():
-            # make sure it wasn't already reduced to zero exponent on prior pass
-            if unit1 not in newunits:
-                continue
-            for unit2 in newunits:
-                if unit1 != unit2:
-                    power = self._REGISTRY._get_dimensionality_ratio(unit1, unit2)
-                    if power:
-                        newunits = newunits.add(unit2, exp / power).remove([unit1])
-                        break
+        units = self._units.copy()
+        new_units = self._get_reduced_units(units)
 
-        return self.ito(newunits)
+        return self.ito(new_units)
 
-    def to_reduced_units(self):
+    def to_reduced_units(self) -> Quantity[_MagnitudeType]:
         """Return Quantity scaled in place to reduced units, i.e. one unit per
         dimension. This will not reduce compound units (intentionally), nor
         can it make use of contexts at this time.
         """
 
-        # can we make this more efficient?
-        newq = copy.copy(self)
-        newq.ito_reduced_units()
-        return newq
+        # shortcuts in case we're dimensionless or only a single unit
+        if self.dimensionless:
+            return self.to({})
+        if len(self._units) == 1:
+            return self
 
-    def to_compact(self, unit=None):
+        units = self._units.copy()
+        new_units = self._get_reduced_units(units)
+
+        return self.to(new_units)
+
+    def to_compact(self, unit=None) -> Quantity[_MagnitudeType]:
         """ "Return Quantity rescaled to compact, human-readable units.
 
         To get output in terms of a different unit, use the unit parameter.
 
 
-        Example
-        -------
+        Examples
+        --------
 
         >>> import pint
         >>> ureg = pint.UnitRegistry()
@@ -772,7 +865,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         ):
             return self
 
-        SI_prefixes = {}
+        SI_prefixes: Dict[int, str] = {}
         for prefix in self._REGISTRY._prefixes.values():
             try:
                 scale = prefix.converter.scale
@@ -783,14 +876,14 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             except Exception:
                 SI_prefixes[0] = ""
 
-        SI_prefixes = sorted(SI_prefixes.items())
-        SI_powers = [item[0] for item in SI_prefixes]
-        SI_bases = [item[1] for item in SI_prefixes]
+        SI_prefixes_list = sorted(SI_prefixes.items())
+        SI_powers = [item[0] for item in SI_prefixes_list]
+        SI_bases = [item[1] for item in SI_prefixes_list]
 
         if unit is None:
-            unit = infer_base_unit(self)
+            unit = infer_base_unit(self, registry=self._REGISTRY)
         else:
-            unit = infer_base_unit(self.__class__(1, unit))
+            unit = infer_base_unit(self.__class__(1, unit), registry=self._REGISTRY)
 
         q_base = self.to(unit)
 
@@ -805,34 +898,34 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             unit_str, unit_power = units[0]
 
         if unit_power > 0:
-            power = math.floor(math.log10(abs(magnitude)) / unit_power / 3) * 3
+            power = math.floor(math.log10(abs(magnitude)) / float(unit_power) / 3) * 3
         else:
-            power = math.ceil(math.log10(abs(magnitude)) / unit_power / 3) * 3
+            power = math.ceil(math.log10(abs(magnitude)) / float(unit_power) / 3) * 3
 
         index = bisect.bisect_left(SI_powers, power)
 
         if index >= len(SI_bases):
             index = -1
 
-        prefix = SI_bases[index]
+        prefix_str = SI_bases[index]
 
-        new_unit_str = prefix + unit_str
+        new_unit_str = prefix_str + unit_str
         new_unit_container = q_base._units.rename(unit_str, new_unit_str)
 
         return self.to(new_unit_container)
 
     # Mathematical operations
-    def __int__(self):
+    def __int__(self) -> int:
         if self.dimensionless:
             return int(self._convert_magnitude_not_inplace(UnitsContainer()))
         raise DimensionalityError(self._units, "dimensionless")
 
-    def __float__(self):
+    def __float__(self) -> float:
         if self.dimensionless:
             return float(self._convert_magnitude_not_inplace(UnitsContainer()))
         raise DimensionalityError(self._units, "dimensionless")
 
-    def __complex__(self):
+    def __complex__(self) -> complex:
         if self.dimensionless:
             return complex(self._convert_magnitude_not_inplace(UnitsContainer()))
         raise DimensionalityError(self._units, "dimensionless")
@@ -1063,6 +1156,14 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         return self.__class__(magnitude, units)
 
+    @overload
+    def __iadd__(self, other: datetime.datetime) -> datetime.timedelta:  # type: ignore[misc]
+        ...
+
+    @overload
+    def __iadd__(self, other) -> Quantity[_MagnitudeType]:
+        ...
+
     def __iadd__(self, other):
         if isinstance(other, datetime.datetime):
             return self.to_timedelta() + other
@@ -1251,11 +1352,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
     __rmul__ = __mul__
 
     def __matmul__(self, other):
-        # Use NumPy ufunc (existing since 1.16) for matrix multiplication
-        if version.parse(NUMPY_VER) >= version.parse("1.16"):
-            return np.matmul(self, other)
-        else:
-            return NotImplemented
+        return np.matmul(self, other)
 
     __rmatmul__ = __matmul__
 
@@ -1432,7 +1529,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             return self
 
     @check_implemented
-    def __pow__(self, other):
+    def __pow__(self, other) -> Quantity[_MagnitudeType]:
         try:
             _to_magnitude(other, self.force_ndarray, self.force_ndarray_like)
         except PintTypeError:
@@ -1484,20 +1581,20 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
                 if getattr(other, "dimensionless", False):
                     exponent = other.to_root_units().magnitude
-                    units = new_self._units ** exponent
+                    units = new_self._units**exponent
                 elif not getattr(other, "dimensionless", True):
                     raise DimensionalityError(other._units, "dimensionless")
                 else:
                     exponent = _to_magnitude(
-                        other, self.force_ndarray, self.force_ndarray_like
+                        other, force_ndarray=False, force_ndarray_like=False
                     )
-                    units = new_self._units ** exponent
+                    units = new_self._units**exponent
 
-            magnitude = new_self._magnitude ** exponent
+            magnitude = new_self._magnitude**exponent
             return self.__class__(magnitude, units)
 
     @check_implemented
-    def __rpow__(self, other):
+    def __rpow__(self, other) -> Quantity[_MagnitudeType]:
         try:
             _to_magnitude(other, self.force_ndarray, self.force_ndarray_like)
         except PintTypeError:
@@ -1508,18 +1605,18 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             if not self.dimensionless:
                 raise DimensionalityError(self._units, "dimensionless")
             new_self = self.to_root_units()
-            return other ** new_self._magnitude
+            return other**new_self._magnitude
 
-    def __abs__(self):
+    def __abs__(self) -> Quantity[_MagnitudeType]:
         return self.__class__(abs(self._magnitude), self._units)
 
-    def __round__(self, ndigits=0):
+    def __round__(self, ndigits: Optional[int] = 0) -> Quantity[int]:
         return self.__class__(round(self._magnitude, ndigits=ndigits), self._units)
 
-    def __pos__(self):
+    def __pos__(self) -> Quantity[_MagnitudeType]:
         return self.__class__(operator.pos(self._magnitude), self._units)
 
-    def __neg__(self):
+    def __neg__(self) -> Quantity[_MagnitudeType]:
         return self.__class__(operator.neg(self._magnitude), self._units)
 
     @check_implemented
@@ -1628,7 +1725,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
     __ge__ = lambda self, other: self.compare(other, op=operator.ge)
     __gt__ = lambda self, other: self.compare(other, op=operator.gt)
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
         # Only cast when non-ambiguous (when multiplicative unit)
         if self._is_multiplicative:
             return bool(self._magnitude)
@@ -1696,7 +1793,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         else:
             return value
 
-    def __array__(self, t=None):
+    def __array__(self, t=None) -> np.ndarray:
         warnings.warn(
             "The unit of the quantity is stripped when downcasting to ndarray.",
             UnitStrippedWarning,
@@ -1704,41 +1801,30 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         )
         return _to_magnitude(self._magnitude, force_ndarray=True)
 
-    def clip(self, first=None, second=None, out=None, **kwargs):
-        minimum = kwargs.get("min", first)
-        maximum = kwargs.get("max", second)
+    def clip(self, min=None, max=None, out=None, **kwargs):
 
-        if minimum is None and maximum is None:
-            raise TypeError("clip() takes at least 3 arguments (2 given)")
-
-        if maximum is None and "min" not in kwargs:
-            minimum, maximum = maximum, minimum
-
-        kwargs = {"out": out}
-
-        if minimum is not None:
-            if isinstance(minimum, self.__class__):
-                kwargs["min"] = minimum.to(self).magnitude
+        if min is not None:
+            if isinstance(min, self.__class__):
+                min = min.to(self).magnitude
             elif self.dimensionless:
-                kwargs["min"] = minimum
+                pass
             else:
                 raise DimensionalityError("dimensionless", self._units)
 
-        if maximum is not None:
-            if isinstance(maximum, self.__class__):
-                kwargs["max"] = maximum.to(self).magnitude
+        if max is not None:
+            if isinstance(max, self.__class__):
+                max = max.to(self).magnitude
             elif self.dimensionless:
-                kwargs["max"] = maximum
+                pass
             else:
                 raise DimensionalityError("dimensionless", self._units)
+        return self.__class__(self.magnitude.clip(min, max, out, **kwargs), self._units)
 
-        return self.__class__(self.magnitude.clip(**kwargs), self._units)
-
-    def fill(self, value):
+    def fill(self: Quantity[np.ndarray], value) -> None:
         self._units = value._units
         return self.magnitude.fill(value.magnitude)
 
-    def put(self, indices, values, mode="raise"):
+    def put(self: Quantity[np.ndarray], indices, values, mode="raise") -> None:
         if isinstance(values, self.__class__):
             values = values.to(self).magnitude
         elif self.dimensionless:
@@ -1748,11 +1834,11 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         self.magnitude.put(indices, values, mode)
 
     @property
-    def real(self):
+    def real(self) -> Quantity[_MagnitudeType]:
         return self.__class__(self._magnitude.real, self._units)
 
     @property
-    def imag(self):
+    def imag(self) -> Quantity[_MagnitudeType]:
         return self.__class__(self._magnitude.imag, self._units)
 
     @property
@@ -1765,7 +1851,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             yield self.__class__(v, self._units)
 
     @property
-    def shape(self):
+    def shape(self) -> Shape:
         return self._magnitude.shape
 
     @shape.setter
@@ -1795,15 +1881,6 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         Wraps np.prod().
         """
-        # TODO: remove after support for 1.16 has been dropped
-        if not HAS_NUMPY_ARRAY_FUNCTION:
-            raise NotImplementedError(
-                "prod is only defined for"
-                " numpy == 1.16 with NUMPY_ARRAY_FUNCTION_PROTOCOL enabled"
-                f" or for numpy >= 1.17 ({np.__version__} is installed)."
-                " Please try setting the NUMPY_ARRAY_FUNCTION_PROTOCOL environment variable"
-                " or updating your numpy version."
-            )
         return np.prod(self, *args, **kwargs)
 
     def __ito_if_needed(self, to_units):
@@ -1812,10 +1889,10 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 
         self.ito(to_units)
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._magnitude)
 
-    def __getattr__(self, item):
+    def __getattr__(self, item) -> Any:
         if item.startswith("__array_"):
             # Handle array protocol attributes other than `__array__`
             raise AttributeError(f"Array protocol attribute {item} not available.")
@@ -1965,7 +2042,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
             self._get_unit_definition(d).reference == offset_unit_dim for d in deltas
         )
 
-    def _ok_for_muldiv(self, no_offset_units=None):
+    def _ok_for_muldiv(self, no_offset_units=None) -> bool:
         """Checks if Quantity object can be multiplied or divided"""
 
         is_ok = True
@@ -1985,7 +2062,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
                 is_ok = False
         return is_ok
 
-    def to_timedelta(self):
+    def to_timedelta(self: Quantity[float]) -> datetime.timedelta:
         return datetime.timedelta(microseconds=self.to("microseconds").magnitude)
 
     # Dask.array.Array ducking
@@ -1999,7 +2076,9 @@ class Quantity(PrettyIPython, SharedRegistryObject):
         return self._magnitude.__dask_keys__()
 
     def __dask_tokenize__(self):
-        return (Quantity, self._magnitude.name, self.units)
+        from dask.base import tokenize
+
+        return (Quantity, tokenize(self._magnitude), self.units)
 
     @property
     def __dask_optimize__(self):
@@ -2077,7 +2156,7 @@ class Quantity(PrettyIPython, SharedRegistryObject):
 _Quantity = Quantity
 
 
-def build_quantity_class(registry):
+def build_quantity_class(registry: BaseRegistry) -> Type[Quantity]:
     class Quantity(_Quantity):
         _REGISTRY = registry
 
