@@ -3,6 +3,7 @@ import functools
 import logging
 import math
 import re
+from contextlib import nullcontext as does_not_raise
 
 import pytest
 
@@ -18,6 +19,7 @@ from pint.testsuite import QuantityTestCase, helpers
 from pint.util import ParserHelper, UnitsContainer
 
 
+# TODO: do not subclass from QuantityTestCase
 class TestUnit(QuantityTestCase):
     def test_creation(self):
         for arg in ("meter", UnitsContainer(meter=1), self.U_("m")):
@@ -97,6 +99,19 @@ class TestUnit(QuantityTestCase):
                 ureg.default_format = spec
                 assert f"{x}" == result, f"Failed for {spec}, {result}"
 
+    def test_unit_formatting_custom(self, monkeypatch):
+        from pint import formatting, register_unit_format
+
+        monkeypatch.setattr(formatting, "_FORMATTERS", formatting._FORMATTERS.copy())
+
+        @register_unit_format("new")
+        def format_new(unit, **options):
+            return "new format"
+
+        ureg = UnitRegistry()
+
+        assert "{:new}".format(ureg.m) == "new format"
+
     def test_ipython(self):
         alltext = []
 
@@ -140,9 +155,34 @@ class TestUnit(QuantityTestCase):
         x = self.U_("m")
         assert 1 / x == self.Q_(1, "1/m")
 
-    def test_unit_pow(self):
-        x = self.U_("m")
-        assert x ** 2 == self.U_("m**2")
+    @pytest.mark.parametrize(
+        ("unit", "power_ratio", "expectation", "expected_unit"),
+        [
+            ("m", 2, does_not_raise(), "m**2"),
+            ("m", dict(), pytest.raises(TypeError), None),
+        ],
+    )
+    def test_unit_pow(self, unit, power_ratio, expectation, expected_unit):
+        with expectation:
+            x = self.U_(unit)
+            assert x**power_ratio == self.U_(expected_unit)
+
+    def test_is_compatible_with(self):
+        unit = self.ureg.Unit("m")
+
+        assert unit.is_compatible_with("m")
+        assert not unit.is_compatible_with("m**2")
+
+        assert unit.is_compatible_with(self.ureg.Unit("m"))
+        assert not unit.is_compatible_with(self.ureg.Unit("m**2"))
+
+        # test other type considered as dimensionless
+        unit = self.ureg.Unit("")
+        assert unit.is_compatible_with(0.5)
+
+    def test_systems(self):
+        unit = self.ureg.Unit("m")
+        assert unit.systems == frozenset({"cgs", "atomic", "Planck", "mks", "SI"})
 
     def test_unit_hash(self):
         x = self.U_("m")
@@ -313,6 +353,9 @@ class TestRegistry(QuantityTestCase):
         assert self.ureg.parse_expression("meter² · second") == self.Q_(
             1, UnitsContainer(meter=2.0, second=1)
         )
+        assert self.ureg.parse_expression("m²·s⁻²") == self.Q_(
+            1, UnitsContainer(meter=2, second=-2)
+        )
         assert self.ureg.parse_expression("meter⁰.⁵·second") == self.Q_(
             1, UnitsContainer(meter=0.5, second=1)
         )
@@ -345,6 +388,12 @@ class TestRegistry(QuantityTestCase):
             kelvin=1, meter=1
         )
 
+    @helpers.requires_numpy
+    def test_parse_with_force_ndarray(self):
+        ureg = UnitRegistry(force_ndarray=True)
+
+        assert ureg.parse_expression("m * s ** -2").units == ureg.m / ureg.s**2
+
     def test_parse_expression_with_preprocessor(self):
         # Add parsing of UDUNITS-style power
         self.ureg.preprocessors.append(
@@ -512,10 +561,10 @@ class TestRegistry(QuantityTestCase):
             return x + y
 
         def gfunc2(x, y):
-            return x ** 2 + y
+            return x**2 + y
 
         def gfunc3(x, y):
-            return x ** 2 * y
+            return x**2 * y
 
         rst = 3.0 * ureg.meter + 1.0 * ureg.centimeter
 
@@ -578,7 +627,7 @@ class TestRegistry(QuantityTestCase):
             g1(1 * ureg.km / ureg.hour, 1 * ureg.hour, 3.0)
         assert (
             g1(3.6 * ureg.km / ureg.hour, 1 * ureg.second)
-            == 1 * ureg.meter / ureg.second ** 2
+            == 1 * ureg.meter / ureg.second**2
         )
 
         with pytest.raises(TypeError):
@@ -736,6 +785,7 @@ class TestCaseInsensitiveRegistry(QuantityTestCase):
         assert ureg.parse_units("j", case_sensitive=False) == UnitsContainer(joule=1)
 
 
+# TODO: do not subclass from QuantityTestCase
 class TestCompatibleUnits(QuantityTestCase):
     def _test(self, input_units):
         gd = self.ureg.get_dimensionality
@@ -830,6 +880,7 @@ class TestRegistryWithDefaultRegistry(TestRegistry):
             d("myk- = 1000 = kilo-")
 
 
+# TODO: remove QuantityTestCase
 class TestConvertWithOffset(QuantityTestCase):
 
     # The dicts in convert_with_offset are used to create a UnitsContainer.
@@ -925,9 +976,6 @@ class TestConvertWithOffset(QuantityTestCase):
         assert ureg._units["canonical"].aliases == (
             "alias1",
             "alias2",
-            "alias3",
-            "alias4",
-            "alias5",
         )
 
         # Define against unknown name
